/* bin.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "bin.hh"
#include "internal.hh"
#include <algorithm>
#include <iostream>
using namespace gltk;

void Bin::on_configure_event()
{
	if (! child) return;
	int width = get_width();
	int height = get_height();
	if (width < reserved_x && height < reserved_y) {
		child->size_allocate(0, 0, 0, 0);
		return;
	}
	width -= reserved_x;
	height -= reserved_y;
	int m;
	int n;
	// width first
	child->get_preferred_width(m, n);
	int w = min_of(width, max_of(n, width - reducible_x));
	child->get_preferred_height_for_width(w, m, n);
	int h = min_of(height, max_of(n, height - reducible_y));
	child_x = (width - w) / 2;
	child_y = (height - h) / 2;
	child->size_allocate(child_x, child_y, w, h);
}

void Bin::get_preferred_width(int & min, int & nat) const
{
	min = reserved_x;
	nat = reserved_x + reducible_x;
	if (child) {
		int m;
		int n;
		child->get_preferred_width(m, n);
		min += m;
		nat += n;
	}
}

void Bin::get_preferred_height(int & min, int & nat) const
{
	min = reserved_y;
	nat = reserved_y + reducible_y;
	if (child) {
		int m;
		int n;
		child->get_preferred_height(m, n);
		min += m;
		nat += n;
	}
}

void Bin::get_preferred_width_for_height(int height, int & min, int & nat) const
{
	min = reserved_x;
	nat = reserved_x + reducible_x;
	if (child) {
		int m;
		int n;
		child->get_preferred_height(m, n);
		int child_h = max_of(0, height - reserved_y);
		if (child_h > n) child_h = max_of(n, n - reducible_y);
		child->get_preferred_width_for_height(child_h, m, n);
		min += m;
		nat += n;
	}
}

void Bin::get_preferred_height_for_width(int width, int & min, int & nat) const
{
	min = reserved_y;
	nat = reserved_y + reducible_y;
	if (child) {
		int m;
		int n;
		child->get_preferred_width(m, n);
		int child_w = max_of(0, width - reserved_x);
		if (child_w > n) child_w = max_of(n, n - reducible_x);
		child->get_preferred_height_for_width(child_w, m, n);
		min += m;
		nat += n;
	}
}

void Bin::on_add(Widget * w)
{
	if (child) throw Error("add to Bin with a child");
	child = w;
}

void Bin::on_remove(Widget * w)
{
	if (child != w) throw Error("removed child not in Bin");
	child = 0;
}

Bin::Bin() :
	child(0),
	reducible_x(0),
	reducible_y(0),
	reserved_x(0),
	reserved_y(0)
{}

void Bin::set_reducible(int red_x, int red_y)
{
	reducible_x = red_x;
	reducible_y = red_y;
}

void Bin::set_reserved(int x, int y)
{
	reserved_x = x;
	reserved_y = y;
}

void Bin::place_bin(int x, int y)
{
	if (child) child->move_allocate(child_x + x, child_y + y);
}

void Bin::set_border_width(unsigned w)
{
	set_reducible(w * 2, w * 2);
}
