/* container.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "container.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <iostream>
using namespace gltk;

void Container::on_set_window(Window * window)
{
	set_children_window(window);
}

void Container::on_realize()
{
	realize_children();
}

void Container::on_unrealize()
{
	unrealize_children();
}

void Container::on_draw()
{
	draw_children();
}

void Container::on_button_event(int button, int state, int x, int y)
{
	if (grabber) grabber->button_event(button, state, x, y);
	else if (Widget * w = get_child(x, y)) {
		w->button_event(button, state, x, y);
	}
	if (! grabber) on_passive_event(x, y);
}

void Container::on_motion_event(int x, int y)
{
	if (grabber) grabber->motion_event(x, y);
	else on_passive_event(x, y);
}

void Container::on_passive_event(int x, int y)
{
	if (Widget * w = get_child(x, y)) {
		if (tracker != w) {
			if (tracker) tracker->entry_event(false);
			tracker = w;
			w->entry_event(true);
		}
		w->passive_event(x, y);
	}
	else if (tracker) {
		tracker->entry_event(false);
		tracker = 0;
	}
}

bool Container::on_key_event(unsigned char key)
{
	if (focus) return focus->keyboard_event(key);
	return false;
}

bool Container::on_special_event(int key)
{
	if (focus) return focus->special_event(key);
	return false;
}

void Container::on_entry_event(bool entered)
{
	if ((! entered) && tracker) {

		tracker->entry_event(false);
		tracker = 0;
	}
}

bool Container::focus_make()
{
	if (focus) {
		if (focus->focus_make()) return true;
		focus = 0;
	}
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) if ((* i)->focus_make()) {
		focus = * i;
		return true;
	}
	return false;
}

bool Container::focus_next()
{
	std::vector<Widget *>::iterator i = child_list.begin();
	if (focus) {
		if (focus->focus_next()) return true;
		// focus is done, find the next one

		do if (i == child_list.end()) { // focus not in the list?
			// end the cycle
			focus = 0;
			return false;
		} while (focus != * (i ++)); // <--- tricky...
	}
	while (i != child_list.end()) {
		if ((* i)->focus_make()) {
			focus = * i;
			return true;
		}
		i ++;
	}
	focus = 0;
	return false;
}

void Container::focus_end()
{
	if (focus) focus->focus_end();
}

Widget * Container::get_child(int x, int y)
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		if ((* i)->contains(x, y)) return * i;
	}
	return 0;
}

void Container::draw_children()
{
	// While Container does not know where the children are located, each child Widget knows its own location.
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		glPushMatrix();
		(* i)->draw();
		glPopMatrix();
	}
}

void Container::set_children_window(Window * window)
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		(* i)->set_window(window);
	}
}

void Container::realize_children()
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		(* i)->realize();
	}
}

void Container::unrealize_children()
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		(* i)->unrealize();
	}
}

Container::Container() :
	grabber(0),
	focus(0),
	tracker(0)
{}

Container::~Container()
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		if ((* i)->get_managed()) delete * i;
	}
}

void Container::add(Widget * child)
{
	child_list.push_back(child);
	child->set_parent(this);
	child->set_window(get_window());
	on_add(child);
	if (get_realized()) child->realize();
}

void Container::remove(Widget * child)
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++)
	if (* i == child) {
		child_list.erase(i);
		if (get_realized()) child->unrealize();
		on_remove(child);
		child->set_window(0);
		child->set_parent(0);
		return;
	}
	throw Error("child not found");
}

void Container::show_all()
{
	show();
	show_all_children(true);
}

void Container::show_all_children(bool recursive)
{
	for (std::vector<Widget *>::iterator i = child_list.begin(); i != child_list.end(); i ++) {
		Container * c = dynamic_cast<Container *>(* i);
		if (recursive && c) c->show_all();
		else (* i)->show();
	}
}

void Container::set_grabber(Widget * child, bool grab)
{
	if (grabber) {
		if (grab) throw("already grabbed!");
		if (grabber != child) throw("child not the grabber!");
		grabber = 0;
		set_grabbing(false);
		return;
	}
	grabber = child;
	set_grabbing(true);
}

bool Container::focus_on(Widget * child)
{
	if (! get_focus()) return false;
	if (focus && focus != child) focus->focus_end();
	focus = child;
	return true;
}
