/* glutwindow.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "glutwindow.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <map>
#include <cassert>
#include <iostream>
namespace gltk {
	class _glutWindow
	{
		static std::map<int, glutWindow *> win_list;

		typedef void (* cb_setter)(glutWindow *, bool);
		static cb_setter cb_set_list[];

		// Relaying Callback functions for GLUT. These could have been static
		// members of Window. But, I don't see why they should be exposed.
		static void display_func();
		static void reshape_func(int width, int height);
		static void mouse_func(int button, int state, int x, int y);
		static void motion_func(int x, int y);
		static void passive_func(int x, int y);
		static void keyboard_func(unsigned char key, int x, int y);
		static void special_func(int key, int x, int y);
		static void entry_func(int state);
	public:
		static void create(glutWindow * w, int parent, int x, int y, int width, int height); // create a GLUT window
		static void create(glutWindow * w, std::string const & name, int width, int height);
		static void destroy(glutWindow * w);
		static void destroy_all();

		// callback
		static void cb_set_display(glutWindow *, bool);
		static void cb_set_reshape(glutWindow *, bool);
		static void cb_set_mouse(glutWindow *, bool);
		static void cb_set_motion(glutWindow *, bool);
		static void cb_set_passive(glutWindow *, bool);
		static void cb_set_keyboard(glutWindow *, bool);
		static void cb_set_special(glutWindow *, bool);
		static void cb_set_entry(glutWindow *, bool);

		static void setup_cb(glutWindow * g, unsigned origin = 0);
	};
}

using namespace gltk;
std::map<int, glutWindow *> _glutWindow::win_list;

_glutWindow::cb_setter _glutWindow::cb_set_list[] = {
	& _glutWindow::cb_set_display,
	& _glutWindow::cb_set_reshape,
	& _glutWindow::cb_set_mouse,
	& _glutWindow::cb_set_motion,
	& _glutWindow::cb_set_passive,
	& _glutWindow::cb_set_keyboard,
	& _glutWindow::cb_set_special,
	& _glutWindow::cb_set_entry
};

void _glutWindow::display_func()
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_display_func();
}

void _glutWindow::reshape_func(int width, int height)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_reshape_func(width, height);
}

void _glutWindow::mouse_func(int button, int state, int x, int y)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_mouse_func(button, state, x, y);
}

void _glutWindow::motion_func(int x, int y)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_motion_func(x, y);
}

void _glutWindow::passive_func(int x, int y)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_passive_func(x, y);
}

void _glutWindow::keyboard_func(unsigned char key, int x, int y)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_keyboard_func(key, x, y);
}

void _glutWindow::special_func(int key, int x, int y)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_special_func(key, x, y);
}

void _glutWindow::entry_func(int state)
{
	glutWindow & w = * win_list[glutGetWindow()];
	w.on_entry_func(state);
}

void _glutWindow::create(glutWindow * g, int p, int x, int y, int w, int h)
{
	assert(! g->win_id);
	if (p) { // create subwindow
		g->win_id = glutCreateSubWindow(p, x, y, w, h);
	}
	else { // create toplevel window
		glutInitWindowSize(w, h);
		glutInitWindowPosition(x, y);
		g->win_id = glutCreateWindow("gltk Window");
	}
	win_list[g->win_id] = g;
	g->on_create();
}

void _glutWindow::create(glutWindow * g, std::string const & name, int w, int h)
{
	glutInitWindowSize(w, h);
	g->win_id = glutCreateWindow(name.c_str());
	g->on_create();
	win_list[g->win_id] = g;
}

void _glutWindow::destroy(glutWindow * g)
{
	win_list.erase(g->win_id);
	g->on_destroy();
	glutDestroyWindow(g->win_id);
	g->win_id = 0;
}

void _glutWindow::cb_set_display(glutWindow * g, bool v)
{
	if (v) glutDisplayFunc(& display_func);
	else glutDisplayFunc(0);
}

void _glutWindow::cb_set_reshape(glutWindow * g, bool v)
{
	if (v) glutReshapeFunc(& reshape_func);
	else glutReshapeFunc(0);
}

void _glutWindow::cb_set_mouse(glutWindow * g, bool v)
{
	if (v) glutMouseFunc(& mouse_func);
	else glutMouseFunc(0);
}

void _glutWindow::cb_set_motion(glutWindow * g, bool v)
{
	if (v) glutMotionFunc(& motion_func);
	else glutMotionFunc(0);
}

void _glutWindow::cb_set_passive(glutWindow * g, bool v)
{
	if (v) glutPassiveMotionFunc(& passive_func);
	else glutPassiveMotionFunc(0);
}

void _glutWindow::cb_set_keyboard(glutWindow * g, bool v)
{
	if (v) glutKeyboardFunc(& keyboard_func);
	else glutKeyboardFunc(0);
}

void _glutWindow::cb_set_special(glutWindow * g, bool v)
{
	if (v) glutSpecialFunc(& special_func);
	else glutSpecialFunc(0);
}

void _glutWindow::cb_set_entry(glutWindow * g, bool v)
{
	if (v) glutEntryFunc(& entry_func);
	else glutEntryFunc(0);
}

void _glutWindow::setup_cb(glutWindow * g, unsigned origin)
{
	for (size_t i = 0; i < 8; i ++) {
		unsigned mask = 1 << i;
		bool f0 = origin & mask;
		bool f1 = g->event_mask & mask;
		if (f1 != f0) (* cb_set_list[i])(g, f1);
	}
}

glutWindow::glutWindow() :
	win_id(0),
	event_mask(EVENT_MASK_NONE)
{
}

glutWindow::~glutWindow()
{
}

void glutWindow::set_events(unsigned mask)
{
	if (win_id) { // created, update event mask
		unsigned om = event_mask;
		event_mask = mask;
		_glutWindow::setup_cb(this, om);
	}
	else event_mask = mask;
}

void glutWindow::add_events(unsigned mask)
{
	if (win_id) { // created, update event mask
		unsigned om = event_mask;
		event_mask |= mask;
		_glutWindow::setup_cb(this, om);
	}
	else event_mask |= mask;
}

void glutWindow::create(int parent_win_id, int x, int y, int width, int height)
{
	_glutWindow::create(this, parent_win_id, x, y, width, height);
	_glutWindow::setup_cb(this);
}

void glutWindow::create(std::string const & name, int width, int height)
{
	_glutWindow::create(this, name, width, height);
	_glutWindow::setup_cb(this);
}

void glutWindow::destroy()
{
	assert(win_id);
	_glutWindow::destroy(this);
}

void glutWindow::set_current()
{
	assert(win_id);
	glutSetWindow(win_id);
}

void glutWindow::move(int x, int y)
{
	glutSetWindow(win_id);
	glutPositionWindow(x, y);
}

void glutWindow::resize(int width, int height)
{
	assert(win_id);
	glutSetWindow(win_id);
	glutReshapeWindow(width, height);
}

void glutWindow::redisplay()
{
	if (win_id) {
		glutSetWindow(win_id);
		glutPostRedisplay();
	}
}
