/* listview.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "listview.hh"
#include "timeout.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <iostream>
#include <algorithm>
#include <cmath>
#include <cassert>
using namespace gltk;

std::string ViewableList::ItemBase::str() const
{
	return "item";
}

int RenderItemStr::get_height()
{
	return glut_get_font_height(font) + 2;
}

void RenderItemStr::render(ViewableList::ItemBase * item, int x, int y, int width, int height)
{
	glColor3f(0, 0, 0);
	glRasterPos2i(x + 1, y + 1 + glut_get_font_ascent(font));
	glut_draw_string(font, item->str(), width - 2);
}

RenderItemStr::RenderItemStr() :
	font(GLUT_BITMAP_HELVETICA_12)
{
}

void ListView::clear_click()
{
	last_click = false;
}

void ListView::on_draw()
{
	int h = get_height();
	int w = get_width();
	glColor3f(1, 1, 1);
	glBegin(GL_QUADS);
	glVertex2i(0, 0);
	glVertex2i(w, 0);
	glVertex2i(w, h);
	glVertex2i(0, h);
	glEnd();
	glut_draw_sunken_box(0, 0, w, h);
	int bw = glut_sunken_box_width();
	int aw = w - bw;
	if (list) {
		glColor3f(0, 0, 0);
		int y = bw;
		if (line_start && line_start + line_visible > (int) list->size()) line_start = max_of(0, int(list->size()) - line_visible);
		line_end = min_of((int) list->size(), line_start + line_visible);
		if (select >= 0 && select < (int) list->size() && select >= line_end) {
			line_start += select + 1 - line_end;
			line_end = select + 1;
		}
		for (int i = line_start; i < line_end; i ++) {
			if (i == select) {
				int sy0 = (i - line_start) * line_height + bw;
				int sy1 = sy0 + line_height;
				if (focused) {
					glColor3f(0.8, 0.9, 1.0);
					glBegin(GL_QUADS);
					glVertex2i(bw, sy0);
					glVertex2i(w - bw, sy0);
					glVertex2i(w - bw, sy1);
					glVertex2i(bw, sy1);
					glEnd();
				}
				else {
					glColor3f(0.6, 0.7, 0.8);
					glBegin(GL_LINE_LOOP);
					glVertex2f(bw + 0.5, sy0 + 0.5);
					glVertex2f(w - bw - 0.5, sy0 + 0.5);
					glVertex2f(w - bw - 0.5, sy1 - 0.5);
					glVertex2f(bw + 0.5, sy1 - 0.5);
					glEnd();
				}
			}
			if (render) render->render(list->get_item(i), bw, y, aw, line_height);
			y += line_height;
		}
		// Should probably calculate these before drawing...
		value = line_start;
		upper = list->size();
		page_increment = line_visible - 1;
		page_size = line_end - line_start;
	}
}

void ListView::on_configure_event()
{
	line_visible = (get_height() - glut_sunken_box_width() * 2) / line_height;
}

void ListView::on_button_event(int button, int state, int x, int y)
{
	if (! focused) {
		if (get_focus()) {
			focused = true;
			queue_draw();
		}
	}
	if (button == GLUT_LEFT_BUTTON) {
		if (state == GLUT_DOWN) {
			int sel = line_start + (int) floor(double(y - 1) / line_height);
			if (sel >= line_start && sel < line_end) {
				if (! last_click) {
					last_click = true;
					gltk::SignalTimeout::connect_once(sigc::mem_fun(this, & ListView::clear_click), 300);
					if (select == sel) return; // no marking change
					select = sel;
				}
				else {
					last_click = false;
					if (select == sel) {
						selected.emit(select);
						select = - 1;
					}
					else select = sel;
				}
			}
			else select = - 1;
			queue_draw();
		}
	}
	else if (button == 3) { // FreeGLUT, wheel up
		if (state != GLUT_DOWN) return;
		if (select < 0) { // no select, scroll page
			if (line_start) line_start --;
			else return;
		}
		else if (select > 0) { // with select, move it
			select --;
			if (select >= 0 && select < line_start) line_start = select;
		}
		else return;
		queue_draw();
	}
	else if (button == 4) { // FreeGLUT, wheel down
		if (state != GLUT_DOWN) return;
		assert(list);
		if (select < 0) { // no select, scroll page
			if (line_start + line_visible < int(list->size())) line_start ++;
			else return;
		}
		else if (select < int(list->size()) - 1) {
			select ++;
			if (select >= line_end) {
				line_start += select - line_end + 1;
			}
		}
		else return;
		queue_draw();
	}
}

bool ListView::on_key_event(unsigned char key)
{
	switch (key) {
	case 0xd:
		if (select >= 0) {
			selected.emit(select);
			select = - 1;
		}
		break;
	case 0x20:
		if (select < 0) {
			select = 0;
			line_start = select;
		}
		else select = - 1;
		break;
	default:
		return false;
	}
	queue_draw();
	return true;
}

bool ListView::on_special_event(int key)
{
	if (! list) return false;
	switch (key) {
	case GLUT_KEY_UP:
		if (select > 0) {
			select --;
		}
		else select = 0;
		break;
	case GLUT_KEY_DOWN:
		if (select >= 0 && select < int(list->size()) - 1) {
			select ++;
		}
		else select = 0;
		break;
	case GLUT_KEY_PAGE_UP:
		if (select < line_visible) select = 0;
		else select -= line_visible;
		break;
	case GLUT_KEY_PAGE_DOWN:
		if (select < 0) select = 0;
		else {
			select += line_visible;
			if (select >= (int) list->size()) select = list->size() - 1;
		}
		break;
	default:
		return false;
	}
	// adjust line_start
	if (select >= 0) {
		if (select < line_start) line_start = select;
		else if (select >= line_end) {
			line_start += select - line_end + 1;
		}
	}
	queue_draw();
	return true;
}

void ListView::on_value_changed()
{
	if (! list) throw Error("No list!");
	int v = value;
	if (v > line_start) {
		line_start = min_of(int(list->size()) - line_visible, v);
		if (line_start < 0) line_start = 0;
		value = line_start;
		if (select >= 0 && select < line_start) select = line_start;
	}
	else if (v < line_start) {
		line_start = v;
		if (select >= 0 && select >= line_start + line_visible) select = line_start + line_visible - 1;
	}
	queue_draw();
}

void ListView::get_preferred_height(int & min, int & nat) const
{
	if (list && render) {
		min = line_height + 4;
		nat = max_of(min, min_of((int(list->size()) + 1) * line_height, 400));
	}
	else {
		min = 20;
		nat = 200;
	}
}

void ListView::get_preferred_width(int & min, int & nat) const
{
	min = 100;
	nat = 200;
}

void ListView::list_change()
{
	queue_draw();
}

ListView::ListView() :
	Adjustment(0, 0, 10, 1, 10, 1),
	list(0),
	select(- 1),
	last_click(false)
{
	line_start = 0;
	line_end = 0;

	lower = 0;
	step_increment = 1;
	page_size = 1;
}

void ListView::set_list(ViewableList * l)
{
	list = l;
}

void ListView::add_render(ViewableList::RenderBase * rndr)
{
	render = rndr;
	line_height = render->get_height();
}

sigc::signal<void, int> & ListView::signal_selected()
{
	return selected;
}
