/* listview.hh
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#ifndef GLTK_LISTVIEW_HH
#define GLTK_LISTVIEW_HH
#include "widget.hh"
#include "focusable.hh"
#include "sigc.hh"
#include "adjustment.hh"
#include <string>
#include <vector>
namespace gltk {
	/** A list of viewable items.
	    This is a simplified take on the model-view design.
	    Items can be any data structures (model) inherit from
	    ItemBase. They are to be rendered (view) by render objects
	    inherit from RenderBase. A tree or list can be used to
	    store data of items. A Widget (such as ListView)can then
	    keep a number of renders to display these data. */
	struct ViewableList
	{
		/// base item type
		struct ItemBase
		{
			virtual std::string str() const; ///< get item in a string
		};
		/// base Render object
		struct RenderBase
		{
			virtual int get_height() = 0; ///< get cell height
			/// render the item into given area
			virtual void render(ItemBase * item, int x, int y, int width, int height) {}
		};

		virtual size_t size() const = 0; ///< size of the list
		virtual ItemBase * get_item(size_t index) = 0; ///< get the index-th item in the list
		sigc::signal<void> changed; ///< signal when the list is changed
	};

	/// render an item as a string
	struct RenderItemStr :
		public ViewableList::RenderBase
	{
		void * font; ///< font used for render
		int get_height();
		void render(ViewableList::ItemBase * item, int x, int y, int width, int height);
		RenderItemStr();
	};

	/** Widget showing and allowing selection from a list of items.
	    The range of view area is an Ajustment that can be controlled
	    by a ScrollBar. */
	class ListView :
		public virtual Widget,
		public virtual Adjustment,
		public virtual Focusable
	{
		sigc::signal<void, int> selected;

		ViewableList * list;
		ViewableList::RenderBase * render;
		int select;

		bool last_click;
		void clear_click();

		void on_draw();
		void on_configure_event();
		void on_button_event(int button, int state, int x, int y);
		bool on_key_event(unsigned char key);
		bool on_special_event(int key);

		void on_value_changed();

		void get_preferred_height(int & min, int & nat) const;
		void get_preferred_width(int & min, int & nat) const;

		// auxiliary
		int line_start;
		int line_end;
		int line_height;
		int line_visible; // size of visible area in lines

		// handler
		void list_change();
	public:
		ListView();
		void set_list(ViewableList * list); ///< set the linked ViewableList
		void add_render(ViewableList::RenderBase * render); ///< add a render to the view (only one is allowed for now)
		sigc::signal<void, int> & signal_selected(); ///< signal for selected item, argument is the index of the item
	};
}
#endif
