/* scale.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "scale.hh"
#include "positiontype.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <sstream>

using namespace gltk;

void Scale::update_label()
{
	std::ostringstream s;
	double v = sl.get_value();
	if (v < 0) {
		s << '-';
		v = - v;
	}
	int t = v;
	s << t;
	v -= t;
	if (digits) {
		s << '.';
		for (int i = 0; i < digits; i ++) {
			v *= 10;
			t = v;
			s << t;
			v -= t;
		}
	}
	set_text(s.str());
	std::istringstream ss(s.str());
	ss >> v;
	sl.set_value(v);
}

void Scale::update_wtxt()
{
	if (! draw_value) return;
	// calculate width needed by the value label
	// assume max width at the bounds
	std::ostringstream s;
	s << int(sl.get_adjustment()->get_lower());
	if (digits) {
		s << '.';
		s << std::string(digits, '0');
	}
	wtxt = text_width(s.str());
	s.str("");
	s << int(sl.get_adjustment()->get_upper());
	if (digits) {
		s << '.';
		s << std::string(digits, '0');
	}
	wtxt = max_of(wtxt, text_width(s.str()));
}

void Scale::on_configure_event()
{
	Bin::on_configure_event();
	if (! draw_value) return;
	int cw = child->get_width();
	int ch = child->get_height();
	int w = get_width();
	int h = get_height();
	switch (pos) {
	case POS_TOP:
		place_bin(0, h - ch);
		break;
	case POS_BOTTOM:
		place_bin(0, 0);
		break;
	case POS_LEFT:
		place_bin(w - cw, 0);
		break;
	case POS_RIGHT:
		place_bin(0, 0);
	}
}

void Scale::on_draw()
{
	draw_children();
	if (draw_value) {
		int w = get_width();
		int h = get_height();
		int tw = text_width();
		int th = text_height();
		int c = sl.get_slider_pos() + sl.get_slider_size() / 2;
		if (sl.get_hori()) {
			if (pos == POS_TOP || pos == POS_BOTTOM) {
				c = max_of(c, tw / 2);
				c = min_of(c, w - tw / 2);
			}
			else {
				c = h / 2;
			}
		}
		else {
			if (pos == POS_LEFT || pos == POS_RIGHT) {
				c = max_of(c, th / 2);
				c = min_of(c, w - th / 2);
			}
			else {
				c = w / 2;
			}
		}
		glColor3f(0, 0, 0);
		switch (pos) {
		case POS_TOP:
			glRasterPos2i(c, 0);
			draw_text(ALIGN_CENTER, ALIGN_START, tw);
			break;
		case POS_BOTTOM:
			glRasterPos2i(c, h);
			draw_text(ALIGN_CENTER, ALIGN_END, tw);
			break;
		case POS_LEFT:
			glRasterPos2i(0, c);
			draw_text(ALIGN_START, ALIGN_CENTER, tw);
			break;
		case POS_RIGHT:
			glRasterPos2i(w, c);
			draw_text(ALIGN_END, ALIGN_CENTER, tw);
			break;
		}
	}
}

void Scale::get_preferred_width(int & min, int & nat) const
{
	Bin::get_preferred_width(min, nat);
	if (draw_value) {
		if (min < wtxt) {
			min = wtxt;
			if (nat < min) nat = min;
		}
	}
}

void Scale::get_preferred_width_for_height(int height, int & min, int & nat) const
{
	Bin::get_preferred_width_for_height(height, min, nat);
	if (draw_value) {
		if (min < wtxt) {
			min = wtxt;
			if (nat < min) nat = min;
		}
	}
}

Scale::Scale(Adjustment * adjust, Orientation orient) :
	sl(adjust, orient),
	pos(POS_TOP),
	digits(3),
	draw_value(true)
{
	update_label();
	update_wtxt();
	set_reserved(0, text_height());
	adjust->signal_value_changed().connect(sigc::mem_fun(this, & Scale::update_label));
	adjust->signal_changed().connect(sigc::mem_fun(this, & Scale::update_wtxt));
	add(sl);
	sl.show();
}

void Scale::set_digits(int dgt)
{
	digits = dgt;
	update_label();
	update_wtxt();
	if (draw_value && (pos == POS_LEFT || pos == POS_RIGHT)) {
		set_reserved(wtxt, 0);
	}
	queue_resize();
}

void Scale::set_draw_value(bool dv)
{
	draw_value = dv;
	update_wtxt();
	if (draw_value) {
		if (pos == POS_TOP || pos == POS_BOTTOM) set_reserved(0, text_height());
		else set_reserved(wtxt, 0);
	}
	else set_reserved(0, 0);
	queue_resize();
}

void Scale::set_value_pos(PositionType p)
{
	pos = p;
	if (draw_value) {
		if (pos == POS_TOP || pos == POS_BOTTOM) set_reserved(0, text_height());
		else set_reserved(wtxt, 0);
		queue_resize();
	}
}
