/* scrollbar.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "scrollbar.hh"
#include "shape.hh"
#include "timeout.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <sstream>
#include <iomanip>
#include <cmath>
#include <cassert>

using namespace gltk;

void Scrollbar::on_configure_event()
{
	int width = get_width();
	int height = get_height();
	if (hori) {
		if (width < 20) {
			up.size_allocate(0, 0, 0, height);
			dn.size_allocate(width, 0, 0, height);
		}
		else {
			up.size_allocate(0, 0, 10, height);
			dn.size_allocate(width - 10, 0, 10, height);
		}
	}
	else {
		if (height < 20) {
			up.size_allocate(0, 0, width, 0);
			dn.size_allocate(0, height, width, 0);
		}
		else {
			up.size_allocate(0, 0, width, 10);
			dn.size_allocate(0, height - 10, width, 10);
		}
	}
}

void Scrollbar::on_draw()
{
	int h = get_height();
	int w = get_width();
	if ((hori ? w : h) < 20) return;
	draw_children();
	double upper = adjust->get_upper();
	double lower = adjust->get_lower();
	double value = adjust->get_value();
	double psize = adjust->get_page_size();
	glBegin(GL_LINES);
	if (hori) {
		glColor3f(0.8, 0.8, 0.8);
		glVertex2f(10, 0.5);
		glVertex2f(w - 10, 0.5);
		glColor3f(1.0, 1.0, 1.0);
		glVertex2f(10, h - 0.5);
		glVertex2f(w - 10, h - 0.5);
	}
	else {
		glColor3f(0.8, 0.8, 0.8);
		glVertex2f(0.5, 10);
		glVertex2f(0.5, h - 10);
		glColor3f(1.0, 1.0, 1.0);
		glVertex2f(w - 0.5, 10);
		glVertex2f(w - 0.5, h - 10);
	}
	glEnd();
	int z = (hori ? w : h) - 20;
	slider_sz = max_of(round_int(psize * z / (upper - lower)), 10);
	slider_z0 = 10;
	if (slider_sz < z) slider_z0 += round_int((value - lower) * (z - slider_sz) / (upper - lower - psize));
	if (hori) glut_draw_raised_box(slider_z0, 0, slider_sz, h);
	else glut_draw_raised_box(0, slider_z0, w, slider_sz);
}

void Scrollbar::on_button_event(int button, int state, int x, int y)
{
	assert(adjust);
	if (button == GLUT_LEFT_BUTTON) {
		if (grabber) {
			assert(! sliding);
			assert(state == GLUT_UP);
			grabber->button_event(button, state, x, y);
		}
		else if (sliding) {
			assert(state == GLUT_UP);
			set_grabbing(false); // release grab
			sliding = false;
		}
		else if (Widget * w = get_child(x, y)) {
			w->button_event(button, state, x, y);
		}
		else if (state == GLUT_DOWN) {
			int z = hori ? x : y;
			if (z < slider_z0) adjust->page_down();
			else if (z >= slider_z0 + slider_sz) adjust->page_up();
			else { // on the slider
				set_grabbing(); // grab it myself
				sliding = true;
				sliding_start = z;
				sliding_vs = adjust->get_value();
			}
		}
		if (grabber && ! held) { // a Button became held
			spin_interval = spin_wait / 2;
			SignalTimeout::connect_once(sigc::mem_fun(this, & Scrollbar::button_held), spin_wait);
			held = true;
			spinning = false;
		}
	}
	else if (button == 3) { // FreeGLUT, wheel up
		if (state == GLUT_DOWN) adjust->step_down();
	}
	else if (button == 4) { // FreeGLUT, wheel down
		if (state == GLUT_DOWN) adjust->step_up();
	}
}

void Scrollbar::on_motion_event(int x, int y)
{
	if (grabber) {
		grabber->motion_event(x, y);
	}
	else {
		assert(sliding);
		int dz = (hori ? x : y) - sliding_start; // mouse displacement
		int rg = (hori ? get_width() : get_height()) - 20 - slider_sz; // available range for y0
		if (! rg) return;
		double vr = adjust->get_upper() - adjust->get_lower() - adjust->get_page_size();
		assert(vr > 0);
		double dv = dz * vr / rg;
		adjust->set_value(sliding_vs + dv);
	}
}

void Scrollbar::button_held()
{
	if (! grabber) { // button released
		held = false;
		return;
	}
	spinning = true;
	if (! adjust) throw Error("No adjust!");
	if (grabber == & up) adjust->step_down();
	else if (grabber == & dn) adjust->step_up();
	else throw Error("Unknown grabber...");
	spin_interval -= (spin_interval / 3);
	spin_interval = max_of(spin_interval, spin_wait / 10);
	SignalTimeout::connect_once(sigc::mem_fun(this, & Scrollbar::button_held), spin_interval);
}

void Scrollbar::pushed_up()
{
	if (! adjust) throw Error("No adjust!");
	adjust->step_down();
}

void Scrollbar::pushed_dn()
{
	if (! adjust) throw Error("No adjust!");
	adjust->step_up();
}

void Scrollbar::get_preferred_width(int & min, int & nat) const
{
	if (hori) {
		min = 30;
		nat = 200;
	}
	else {
		min = 16;
		nat = 16;
	}
}

void Scrollbar::get_preferred_height(int & min, int & nat) const
{
	if (hori) {
		min = 16;
		nat = 16;
	}
	else {
		min = 30;
		nat = 200;
	}
}


Scrollbar::Scrollbar(Adjustment * adj, Orientation orient) :
	adjust(adj),
	hori(orient == ORIENTATION_HORIZONTAL),
	up(new Shape((hori ? Shape::LEFT_TRIANGLE : Shape::UP_TRIANGLE), 15, 8)),
	dn(new Shape((hori ? Shape::RIGHT_TRIANGLE : Shape::DOWN_TRIANGLE), 15, 8)),
	spin_wait(700),
	held(false),
	spinning(false),
	sliding(false)
{
	add(& up);
	add(& dn);
	// reduced box_width for more compact arrows
	up.set_reserved(5, 5);
	dn.set_reserved(5, 5);
	up.set_box_width(2);
	dn.set_box_width(2);
	up.signal_pushed().connect(sigc::mem_fun(this, & Scrollbar::pushed_up));
	dn.signal_pushed().connect(sigc::mem_fun(this, & Scrollbar::pushed_dn));
}
