/* slider.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "slider.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <cassert>

using namespace gltk;

void Slider::on_draw()
{
	int w = get_width();
	int h = get_height();

	int z = hori ? w : h;
	// draw shade
	if (adjust) {
		double upper = adjust->get_upper() - adjust->get_page_size();
		double lower = adjust->get_lower();
		double value = adjust->get_value();
		if (upper > lower) {
			slider_z0 = (value - lower) * (z - slider_sz) / (upper - lower);
			glBegin(GL_QUADS);
			glColor3f(0.8, 0.8, 1.0);
			if (hori) {
				glVertex2i(0, (h - slider_sz) / 2);
				glVertex2i(slider_z0, (h - slider_sz) / 2);
				glVertex2i(slider_z0, (h + slider_sz) / 2);
				glVertex2i(0, (h + slider_sz) / 2);
			}
			else {
				glVertex2i((w - slider_sz) / 2, 0);
				glVertex2i((w - slider_sz) / 2, slider_z0);
				glVertex2i((w + slider_sz) / 2, slider_z0);
				glVertex2i((w + slider_sz) / 2, 0);
			}
			glEnd();
		}
		else slider_z0 = - 1;
	}

	// draw center line
	int mz = (hori ? h : w) / 2;
	glBegin(GL_LINES);
	glColor3f(0.7, 0.7, 0.7);
	if (hori) {
		glVertex2f(slider_sz / 2.0, mz - 0.5);
		glVertex2f(w - slider_sz / 2.0, mz - 0.5);
	}
	else {
		glVertex2f(mz - 0.5, slider_sz / 2.0);
		glVertex2f(mz - 0.5, h - slider_sz / 2.0);
	}
	glColor3f(1.0, 1.0, 1.0);
	if (hori) {
		glVertex2f(slider_sz / 2.0, mz + 0.5);
		glVertex2f(w - slider_sz / 2.0, mz + 0.5);
	}
	else {
		glVertex2f(mz + 0.5, slider_sz / 2.0);
		glVertex2f(mz + 0.5, h - slider_sz / 2.0);
	}
	glEnd();
	if (adjust && slider_z0 >= 0) {
		glBegin(GL_QUADS);
		glColor3f(0.9, 0.9, 0.9);
		if (hori) {
			glVertex2i(slider_z0, (h - slider_sz) / 2);
			glVertex2i(slider_z0 + slider_sz, (h - slider_sz) / 2);
			glVertex2i(slider_z0 + slider_sz, (h + slider_sz) / 2);
			glVertex2i(slider_z0, (h + slider_sz) / 2);
		}
		else {
			glVertex2i((w - slider_sz) / 2, slider_z0);
			glVertex2i((w - slider_sz) / 2, slider_z0 + slider_sz);
			glVertex2i((w + slider_sz) / 2, slider_z0 + slider_sz);
			glVertex2i((w + slider_sz) / 2, slider_z0);
		}
		glEnd();
		if (hori) glut_draw_raised_box(slider_z0, (h - slider_sz) / 2, slider_sz, slider_sz);
		else glut_draw_raised_box((w - slider_sz) / 2, slider_z0, slider_sz, slider_sz);
	}
}

void Slider::on_button_event(int button, int state, int x, int y)
{
	assert(adjust);
	if (button == GLUT_LEFT_BUTTON) {
		if (sliding) {
			assert(state == GLUT_UP);
			set_grabbing(false); // release grab
			sliding = false;
		}
		else if (state == GLUT_DOWN && slider_z0 >= 0) {
			int z = hori ? x : y;
			if (z < slider_z0) adjust->page_down();
			else if (z >= slider_z0 + slider_sz) adjust->page_up();
			else { // on the slider
				set_grabbing(); // grab it myself
				sliding = true;
				sliding_start = z;
				sliding_vs = adjust->get_value();
			}
		}
	}
	else if (button == 3) { // FreeGLUT, wheel up
		if (state == GLUT_DOWN) adjust->step_down();
	}
	else if (button == 4) { // FreeGLUT, wheel down
		if (state == GLUT_DOWN) adjust->step_up();
	}
}

void Slider::on_motion_event(int x, int y)
{
	assert(sliding);
	int dz = (hori ? x : y) - sliding_start; // mouse displacement
	int rg = (hori ? get_width() : get_height()) - slider_sz; // available range for y0
	if (! rg) return;
	double vr = adjust->get_upper() - adjust->get_lower() - adjust->get_page_size();
	assert(vr > 0);
	double dv = dz * vr / rg;
	adjust->set_value(sliding_vs + dv);
}

void Slider::get_preferred_width(int & min, int & nat) const
{
	if (hori) {
		min = 30;
		nat = 120;
	}
	else {
		min = slider_sz;
		nat = slider_sz;
	}
}

void Slider::get_preferred_height(int & min, int & nat) const
{
	if (hori) {
		min = slider_sz;
		nat = slider_sz;
	}
	else {
		min = 30;
		nat = 120;
	}
}

Slider::Slider(Adjustment * adj, Orientation ori) :
	Orientable(ori),
	adjust(adj),
	slider_sz(16),
	sliding(false)
{
	if (! adjust) adjust = new Adjustment(0, 0, 10);
	adjust->signal_value_changed().connect(sigc::mem_fun(this, & Slider::queue_draw));
}

Adjustment * Slider::get_adjustment()
{
	return adjust;
}

void Slider::set_value(double value)
{
	assert(adjust);
	adjust->set_value(value);
}

double Slider::get_value() const
{
	return adjust->get_value();
}

int Slider::get_slider_pos() const
{
	return slider_z0;
}

int Slider::get_slider_size() const
{
	return slider_sz;
}

sigc::signal<void> & Slider::signal_value_changed()
{
	assert(adjust);
	return adjust->signal_value_changed();
}
