/* widget.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "window.hh"
#include "container.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <iostream>
#include <cassert>
using namespace gltk;

Widget::Widget() :
	parent(0),
	window(0),
	managed(false),
	grabbing(false),
	shown(false),
	alloc_x(- 1),
	alloc_y(- 1),
	alloc_w(- 1),
	alloc_h(- 1)
{}

Widget::~Widget()
{
}

void Widget::get_preferred_width(int & min, int & nat) const
{
	min = 0;
	nat = 0;
}

void Widget::get_preferred_height(int & min, int & nat) const
{
	min = 0;
	nat = 0;
}

void Widget::get_preferred_height_for_width(int width, int & min, int & nat) const
{
	get_preferred_height(min, nat);
}

void Widget::get_preferred_width_for_height(int height, int & min, int & nat) const
{
	get_preferred_width(min, nat);
}

bool Widget::focus_make()
{
	return false; // can not take focus by default
}

bool Widget::focus_next()
{
	focus_end();  // automatically ends the focus
	return false; // can not rotate focus by default
}

bool Widget::get_focus()
{
	if (! parent) return false;
	return parent->focus_on(this);
}

void Widget::set_parent(Container * p)
{
	parent = p;
	on_set_parent(p);
}

void Widget::set_managed(bool v)
{
	managed = v;
}

bool Widget::get_managed() const
{
	return managed;
}

void Widget::realize()
{
	configured = false;
	on_realize();
}

void Widget::unrealize()
{
	on_unrealize();
}

void Widget::size_allocate(int x, int y, int width, int height)
{
	alloc_x = x;
	alloc_y = y;
	if (alloc_w != width || alloc_h != height) {
		alloc_w = width;
		alloc_h = height;
	}
	configured = true;
	on_configure_event();
}

void Widget::move_allocate(int x, int y)
{
	assert(configured);
	alloc_x = x;
	alloc_y = y;
	on_configure_event();
}

bool Widget::contains(int x, int y) const
{
	return x >= alloc_x && y >= alloc_y && x - alloc_x < alloc_w && y - alloc_y < alloc_h;
}

void Widget::draw()
{
	assert(get_window()->get_id() == glutGetWindow());
	if (alloc_w > 0 && alloc_h > 0) {
		if (alloc_x > 0 || alloc_y > 0) glTranslatef(alloc_x, alloc_y, 0);
		on_draw();
		/*
		glColor3ub(255, 0, 0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(0.5, 0.5);
		glVertex2f(alloc_w - 0.5, 0.5);
		glVertex2f(alloc_w - 0.5, alloc_h - 0.5);
		glVertex2f(0.5, alloc_h - 0.5);
		glEnd();
		*/
	}
}

void Widget::show()
{
	shown = true;
	on_show();
}

void Widget::hide()
{
	on_hide();
	shown = false;
}

void Widget::make_focus()
{
	if (focus_make()) {
		if (! get_focus()) focus_end();
	}
}

void Widget::button_event(int button, int state, int x, int y)
{
	if (alloc_x > 0 || alloc_y > 0) {
		x -= alloc_x;
		y -= alloc_y;
	}
	on_button_event(button, state, x, y);
}

void Widget::motion_event(int x, int y)
{
	if (alloc_x > 0 || alloc_y > 0) {
		x -= alloc_x;
		y -= alloc_y;
	}
	on_motion_event(x, y);
}

bool Widget::keyboard_event(unsigned char key)
{
	return on_key_event(key);
}

bool Widget::special_event(int key)
{
	return on_special_event(key);
}

void Widget::passive_event(int x, int y)
{
	if (alloc_x > 0 || alloc_y > 0) {
		x -= alloc_x;
		y -= alloc_y;
	}
	on_passive_event(x, y);
}

void Widget::entry_event(bool entered)
{
	on_entry_event(entered);
}

void Widget::set_grabbing(bool v)
{
	if (grabbing == v) return;
	grabbing = v;
	if (parent) parent->set_grabber(this, v);
}

bool Widget::get_realized() const
{
	return window && window->is_realized();
}

Window * Widget::get_window() const
{
	return window;
}

void Widget::set_window(Window * w)
{
	window = w;
	on_set_window(w);
}

Allocation Widget::get_allocation() const
{
	Allocation a = {alloc_x, alloc_y, alloc_w, alloc_h};
	return a;
}

void Widget::set_size_request(int width, int height)
{
	// not implemented
}

void Widget::get_absolute(int & x, int & y) const
{
	assert(window);
	if (parent && parent->window == window) {
		parent->get_absolute(x, y);
		assert(alloc_x >= 0 &&  alloc_y >= 0);
		x += alloc_x;
		y += alloc_y;
	}
	else {
		x = alloc_x;
		y = alloc_y;
	}
}

void Widget::queue_draw()
{
	if (window) window->refresh();
}

void Widget::queue_resize()
{
	// this can be costly.
	// may better to just ask parent.
	if (window) window->reconfigure();
}
